/***************************************************************************
                           info.c  -  description
                             -------------------
    begin             : 03.11.2010
    last modified     : 14.11.2015    
    copyright         : (C) 2010 - 2015 by MHS-Elektronik GmbH & Co. KG, Germany
    author            : Klaus Demlehner, klaus@mhs-elektronik.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software, you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation.             *
 *                                                                         *
 ***************************************************************************/

/**
    Library to talk to Tiny-CAN devices. You find the latest versions at
       http://www.tiny-can.com/
**/
#include "global.h"
#include <string.h>
#include <stdio.h>
#include "util.h"
#include "tar_drv.h"
#include "errors.h"
#include "var_obj.h"  // Value Typen
#include "info.h"


#define IT_BYTE     0x01
#define IT_UBYTE    0x02
#define IT_HBYTE    0x0A
#define IT_WORD     0x03
#define IT_UWORD    0x04
#define IT_HWORD    0x0C
#define IT_LONG     0x05
#define IT_ULONG    0x06
#define IT_HLONG    0x0E
#define IT_REVISION 0x40     // Software Version
#define IT_DATE     0x41     // Datum
#define IT_STRING   0x80


#define INFO_TAB_SIZE 23

struct TInfoTabItem
  {
  uint16_t Index;
  uint16_t Maske;
  char *Bezeichner;
  };

static const struct TInfoTabItem InfoTab[INFO_TAB_SIZE] = {
  {0x1000, 0xFFFF, "ID"},
  {0x1001, 0xFFFF, "ID String"},
  {0x1002, 0xFFFF, "Version"},
  {0x1003, 0xFFFF, "Version String"},
  {0x1004, 0xFFFF, "Autor"},
  {0x1005, 0xFFFF, "Optionen"},
  {0x1006, 0xFFFF, "Snr"},
  {0x0002, 0xFFFF, "Bios ID String"},
  {0x0001, 0xFFFF, "Hardware ID String"},
  {0x0000, 0xFFFF, "Hardware Snr"},
  {0x0003, 0xFFFF, "Hardware Revision"},
  {0x0004, 0xFFFF, "Fertigungsdatum"},
  {0x000A, 0xFFFF, "Ident-String"},
  {0x8000, 0xFFFF, "Anzahl CAN Interfaces"},
  {0x8001, 0xFFFF, "CAN Features Flags"},
  {0x8010, 0xFFFF, "Treiber"},
  {0x8020, 0xFFFF, "Opto"},
  {0x8030, 0xFFFF, "Term"},
  {0x8040, 0xFFFF, "HighSpeed"},
  {0x8050, 0xFFFF, "Anzahl Interval Puffer"},
  {0x8060, 0xFFFF, "Anzahl Filter"},
  {0x8100, 0xFFFF, "Anzahl I2C Interfaces"},
  {0x8200, 0xFFFF, "Anzahl SPI Interfaces"}};

static const unsigned char InfoVarTypeTranslate[] =
  {0x00,      // 0x00
   VT_BYTE,   // 0x01 BYTE
   VT_UBYTE,  // 0x02 UBYTE
   VT_WORD,   // 0x03 WORD
   VT_UWORD,  // 0x04 UWORD
   VT_LONG,   // 0x05 LONG
   VT_ULONG,  // 0x06 ULONG
   0x00,      // 0x07
   0x00,      // 0x08
   0x00,      // 0x09
   VT_HBYTE,  // 0x0A HBYTE
   0x00,      // 0x0B
   VT_HWORD,  // 0x0C HWORD
   0x00,      // 0x0D
   VT_HLONG,  // 0x0E HLONG
   0x00};     // 0x0F


static uint32_t mhs_pow10(uint32_t n)
{
uint32_t out;

out = 1;
for (; n; n--)
  {
  out *= 10;
  }
return(out);
}


static uint32_t GetBCD(uint32_t in, uint32_t len)
{
uint32_t out, exp;

out = 0;
exp = 0;
for (;len; len--)
  {
  out += ((in & 0x0F) * mhs_pow10(exp));
  exp++;
  in = in >> 4; 
  }
return(out);  
}


void InfoVarDestroy(struct TInfoVarList **liste)
{
struct TInfoVarList *l, *h;

if (!liste)
  return;
l = *liste;
while (l)
  {
  h = l->Next;
  safe_free(l->Data);
  safe_free(l);
  l = h;
  }
*liste = NULL;
}


static struct TInfoVarList *InfoVarAdd(struct TInfoVarList **liste)
{
struct TInfoVarList *l;

l = *liste;
if (!l)
  {
  // Liste ist leer
  l = (struct TInfoVarList *)mhs_calloc(1, sizeof(struct TInfoVarList));
  *liste = l;
  }
else
  { // Neues Element anhngen
  while (l->Next != NULL) l = l->Next;
  l->Next = (struct TInfoVarList *)mhs_calloc(1, sizeof(struct TInfoVarList));
  l = l->Next;
  if (!l)
    return(NULL);        // Nicht gengend Arbetsspeicher
  }
l->Next = NULL;
return(l);
}


static char *CreateInfoString(struct TInfoVarList *info)
{
int32_t size, n;
uint32_t value, tag, monat, jahr, major,  minor, revision;
char *str, *p;
struct TInfoVarList *i;

size = 0;
for (i = info; i; i = i->Next)
  {
  if (i->IndexString)
    size += ((int32_t)strlen(i->IndexString) + 1);
  else
    size += 7;
  switch (i->Type)
    {
    case IT_BYTE   : {
                     size += 5;
                     break;
                     }
    case IT_UBYTE  : {
                     size += 4;
                     break;
                     }
    case IT_HBYTE  : {
                     size += 5;
                     break;
                     }
    case IT_WORD   : {
                     size += 7;
                     break;
                     }
    case IT_UWORD  : {
                     size += 6;
                     break;
                     }
    case IT_HWORD  : {
                     size += 7;
                     break;
                     }
    case IT_LONG   : {
                     size += 12;
                     break;
                     }
    case IT_ULONG  : {
                     size += 11;
                     break;
                     }
    case IT_HLONG  : {
                     size += 11;
                     break;
                     }
    case IT_REVISION :      // Software Version
                     {
                     size += 12;
                     break;
                     }
    case IT_DATE   : {      // Datum
                     size += 11;
                     break;
                     }                     
    case IT_STRING : {
                     i->Data[i->Size] = '\0';
                     size += ((int32_t)strlen(i->Data) + 1);
                     break;
                     }
    }
  }

if (!size)
  return(NULL);
str = (char *)mhs_malloc(size+1);
str[size++] = '\0';
p = str;

for (; info; info = info->Next)
  {
  if (info->IndexString)
    n = safe_sprintf(p, size, "%s=", info->IndexString);
  else
    n = safe_sprintf(p, size, "%#04X=", info->Index);
  p += n;
  size -= n;
  switch (info->Type)
    {
    case IT_BYTE   : {
                     n = safe_sprintf(p, size, "%d;", *((signed char*)(info->Data)));
                     break;
                     }
    case IT_UBYTE  : {
                     n = safe_sprintf(p, size, "%u;", *((unsigned char *)(info->Data)));
                     break;
                     }
    case IT_HBYTE  : {
                     n = safe_sprintf(p, size, "%#02X;", *((unsigned char *)(info->Data)));
                     break;
                     }
    case IT_WORD   : {
                     n = safe_sprintf(p, size, "%d;", *((signed short *)(info->Data)));
                     break;
                     }
    case IT_UWORD  : {
                     n = safe_sprintf(p, size, "%u;", *((uint16_t *)(info->Data)));
                     break;
                     }
    case IT_HWORD  : {
                     n = safe_sprintf(p, size, "%#04X;", *((uint16_t *)(info->Data)));
                     break;
                     }
    case IT_LONG   : {
                     //p += sprintf(p, "%ld;", *((int32_t *)(info->Data)));
                     n = safe_sprintf(p, size, "%d;", *((int32_t *)(info->Data)));
                     break;
                     }
    case IT_ULONG  : {
                     //p += sprintf(p, "%lu;", *((uint32_t *)(info->Data)));
                     n = safe_sprintf(p, size, "%u;", *((uint32_t *)(info->Data)));
                     break;
                     }
    case IT_HLONG  : {
                     //p += sprintf(p, "%#08lX;", *((uint32_t *)(info->Data)));
                     n = safe_sprintf(p, size, "%#08X;", *((uint32_t *)(info->Data)));
                     break;
                     }
    case IT_REVISION :
                     {
                     value = *((uint32_t *)(info->Data));
                     major = value / 10000;
                     value = value % 10000;  
                     minor = value / 100;                   
                     revision = value % 100;                   
                     n = safe_sprintf(p, size, "%u.%u.%u;", major, minor, revision); 
                     break; 
                     }
    case IT_DATE   : {
                     value = *((uint32_t *)(info->Data));
                     tag = GetBCD(value & 0xFF, 2);
                     monat = GetBCD((value >> 8) & 0xFF, 2);  
                     jahr = GetBCD((value >> 16) & 0xFFFF, 4);
                     n = safe_sprintf(p, size, "%02u.%02u.%4u;", tag, monat, jahr);
                     break; 
                     }                     
    case IT_STRING : {
                     info->Data[info->Size] = '\0';
                     n = safe_sprintf(p, size, "%s;", info->Data);
                     break;
                     }
    default        : n = 0;
    }
  p += n;
  size -= n;
  }
return(str);
}


static const char *GetIndexString(uint16_t index)
{
int32_t i;
const struct TInfoTabItem *item;

for (i = 0; i < INFO_TAB_SIZE; i++)
  {
  item = &InfoTab[i];
  if (item->Index == (index & item->Maske))
    return(item->Bezeichner);
  }
return(NULL);
}


struct TInfoVarList *GetInfoByIndex(TCanDevice *dev, uint16_t index)
{
struct TInfoVarList *info;

for (info = dev->InfoVars; info; info = info->Next)
  {
  if (info->Index == index)
    return(info);
  }
return(NULL);
}


uint32_t GetInfoValueAsULong(struct TInfoVarList *info)
{
uint32_t value;

value = 0;
if (!info)
  return(value);
switch (info->Type)
  {
  case IT_BYTE   :
  case IT_UBYTE  :
  case IT_HBYTE  : {
                   value = *((unsigned char *)(info->Data));
                   break;
                   }
  case IT_WORD   :
  case IT_UWORD  :
  case IT_HWORD  : {
                   value = *((uint16_t *)(info->Data));
                   break;
                   }
  case IT_LONG   :
  case IT_ULONG  :
  case IT_HLONG  : {
                   value = *((uint32_t *)(info->Data));
                   break;
                   }
  }

return(value);
}


char *GetInfoValueAsString(struct TInfoVarList *info)
{
if (!info)
  return(NULL);
if (info->Type != IT_STRING)
  return(NULL);
info->Data[info->Size] = '\0';
return(info->Data);
}


int32_t HwReadInfoVar(TCanDevice *dev)
{
int32_t res;
struct TInfoVarList *info;
struct TTarInfoVar info_tmp;

safe_free(dev->DrvInfoHwStr);
InfoVarDestroy(&dev->InfoVars);
if (!(res = TARInfoStartRead(dev)))
  {
  while ((res = TARInfoGetNext(dev, &info_tmp)) > 0)
    {
    if (!(info = InfoVarAdd(&dev->InfoVars)))
      {
      res = -1;
      break;
      }
    info->Size = info_tmp.Size;
    info->Type = info_tmp.Type;
    info->Index = info_tmp.Index;
    info->IndexString = GetIndexString(info_tmp.Index);
    if (info_tmp.Size)
      {
      info->Data = mhs_calloc(1, (info_tmp.Size + 1));
      if (!info->Data)
        {
        res = -1;
        break;
        }
      memcpy(info->Data, info_tmp.Data, info_tmp.Size);
      }
    }
  }
if (res < 0)   // Fehler
  InfoVarDestroy(&dev->InfoVars);
else
  dev->DrvInfoHwStr = CreateInfoString(dev->InfoVars);
return(res);
}


int32_t GetHardwareInfo(TCanDevice *dev, struct TInfoVar **info_var)
{
struct TInfoVarList *info;
struct TInfoVar *out;
uint32_t size, i;
int32_t cnt;

if (!info_var)
  return(ERR_PARAM);
//*info_var = NULL;
out = NULL;
// Anzahl der Eintrge ermitteln
cnt = 0;
for (info = dev->InfoVars; info; info = info->Next)
  cnt++;
if (cnt > 0)
  {
  i = 0;
  out = (struct TInfoVar*)mhs_calloc(cnt, sizeof(struct TInfoVar));
  for (info = dev->InfoVars; info; info = info->Next)
    {
    out[i].Key = info->Index;    
    if (info->Type == IT_STRING)
      out[i].Type = VT_STRING;
    else if (info->Type == IT_REVISION)      // Software Version
      out[i].Type = VT_REVISION;
    else if (info->Type == IT_DATE)          // Datum
      out[i].Type = VT_DATE;
    else
      out[i].Type = InfoVarTypeTranslate[(info->Type & 0x000F)];
    size = info->Size;
    out[i].Size = size;
    if ((size) && (size <= 255))
      {
      memcpy(out[i].Data, info->Data, size);
      if (info->Type == IT_STRING)
        out[i].Data[size] = '\0';
      }
    i++;
    }
  }
*info_var = out;
return(cnt);
}
