/***************************************************************************
                         log.c  -  description
                             -------------------
    begin             : 23.03.2008
    copyright         : (C) 2008 by MHS-Elektronik GmbH & Co. KG, Germany
    author            : Klaus Demlehner, klaus@mhs-elektronik.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software, you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation.             *
 *                                                                         *
 ***************************************************************************/

/**
    Library to talk to Tiny-CAN devices. You find the latest versions at
       http://www.tiny-can.com/
**/
#include "global.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "can_types.h"
#include "drv_types.h"
#include "usb_hlp.h"
#include "util.h"
//#include "var.h"
#include "log.h"


/**************************************************************************/
/*                         V A R I A B L E N                              */
/**************************************************************************/
static FILE *LogFile;
static DRV_LOCK_TYPE LogLock;
static uint32_t LogCreated;
static uint32_t LogFlags;

uint32_t LogEnable;


/**************************************************************************/
/*                        C O N S T A N T E N                             */
/**************************************************************************/
#define DRV_STATUS_MAX 9

static const char *DrvStatusStrings[] = {
  "NOT_LOAD",       // 0  Die Treiber DLL wurde noch nicht geladen
  "NOT_INIT",       // 1  Treiber noch nicht Initialisiert (Funktion "CanInitDrv" noch nicht aufgerufen)
  "INIT",           // 2  Treiber erfolgrich Initialisiert
  "PORT_NOT_OPEN",  // 3  Die Schnittstelle wurde nicht geffnet
  "PORT_OPEN",      // 4  Die Schnittstelle wurde geffnet
  "DEVICE_FOUND",   // 5  Verbindung zur Hardware wurde Hergestellt
  "CAN_OPEN",       // 6  Device wurde geffnet und erfolgreich Initialisiert
  "CAN_RUN_TX",     // 7  CAN Bus RUN nur Transmitter (wird nicht verwendet !)
  "CAN_RUN",        // 8  CAN Bus RUN
  "UNBEKANNT"};

#define CAN_STATUS_MAX 5

static const char *CanStatusStrings[] = {
  "OK",          // 0  Ok
  "ERROR",       // 1  CAN Error
  "WARNING",     // 2  Error warning
  "PASSIV",      // 3  Error passiv
  "BUS_OFF",     // 4  Bus Off
  "UNBEKANNT"};  // 5  Status Unbekannt

#define FIFO_STATUS_MAX 4

static const char *FifoStatusStrings[] = {
  "OK",                 // 0  Fifo-Status: Ok
  "HW_OVERRUN",         // 1  Fifo-Status: Hardware Fifo berlauf
  "SW_OVERRUN",         // 2  Fifo-Status: Software Fifo berlauf
  "HW_SW_OVERRUN",      // 3  Fifo-Status: Hardware & Software Fifo berlauf
  "UNBEKANNT"};         // 4  Fifo-Status: Unbekannt


/**************************************************************************/
/*                        F U N K T I O N E N                             */
/**************************************************************************/

/*
******************** LogInit ********************
Funktion  : Initialisiert den Treiber

Eingaben  : keine

Ausgaben  : keine

Call's    : keine
*/
void LogInit(void)
{
LogFile = NULL;
LogEnable = 0;
LogCreated = 0;
}


/*
******************** LogCreate ********************
Funktion  :

Eingaben  : keine

Ausgaben  : keine

Call's    : DRV_LOCK_INIT
*/
int32_t LogCreate(void)
{
if (!LogCreated)
  {
  DRV_LOCK_INIT(&LogLock);
  LogCreated = 1;
  }
return(0);
}

/*
******************** LogDestroy ********************
Funktion  :

Eingaben  : keine

Ausgaben  : keine

Call's    : DRV_LOCK_DESTROY
*/
void LogDestroy(void)
{
if (LogCreated)
  {
  LogClose();
  DRV_LOCK_DESTROY(&LogLock);
  LogCreated = 0;
  }
}


/*
******************** LogOpen ********************
Funktion  : Log File ffnen

Eingaben  : filename => Log Filename

Ausgaben  : result => 0 = kein Fehler
                      -1 = Fehler beim ffnen/anlegen des Log Files

Call's    : fopen, LogClose
*/
int32_t LogOpen(char *filename, uint32_t flags)
{
if (!LogCreated)
  return(-1);
LogClose();
DRV_LOCK_ENTER(&LogLock);
LogFlags = flags;
//LogFile = fopen(filename, "a"); //O_WRONLY|O_APPEND|O_CREAT);
#if _MSC_VER >= 1500
if (fopen_s(&LogFile, filename, "w+"))
  LogFile = NULL;
#else
LogFile = fopen(filename, "w+");
#endif
//strncpy(LogFileName, filename, MAX_FILENAME_SIZE);
DRV_LOCK_LEAVE(&LogLock);
if (LogFile)
  {
  LogEnable = 1;
  return(0);
  }
else
  return(-1);
}


/*
******************** LogClose ********************
Funktion  : Log File schlieen

Eingaben  : keine

Ausgaben  : keine

Call's    : fclose, fflush, DRV_LOCK_ENTER, DRV_LOCK_LEAVE
*/
void LogClose(void)
{
if (!LogCreated)
  return;
DRV_LOCK_ENTER(&LogLock);
LogEnable = 0;
if (LogFile)
  {
  fflush(LogFile);
  (void)fclose(LogFile);
  }
LogFile = NULL;
//LogFileName[0] = (char)0;
DRV_LOCK_LEAVE(&LogLock);
}




/*
******************** LogSet ********************
Funktion  :

Eingaben  : keine

Ausgaben  : keine

Call's    : DRV_LOCK_DESTROY
*/
/*void LogSet(uint32_t log_flags)
{
if (LogEnable)
  {
  if (log_flags & LOG_LOCK)
    DRV_LOCK_ENTER(&LogLock);
  else if (log_flags & LOG_UNLOCK)
    DRV_LOCK_LEAVE(&LogLock);
  }
}*/


/*
******************** LogPrintf ********************
Funktion  : Einen "printf" formatierten String in das Log File schreiben

Eingaben  : log_type =>
            format_str => printf Format String
            ... => Printf Parameter

Ausgaben  : keine

Call's    : va_start, vfprintf, va_end, fflush
*/
void LogPrintf(uint32_t log_type, uint32_t log_flags, const char *format_str, ...)
{
va_list ap;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(&LogLock);
  va_start(ap, format_str);
  vfprintf(LogFile, format_str, ap);
  va_end(ap);
  fprintf(LogFile, "\n");
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(&LogLock);
  }
}


static void LogWriteMessages(struct TCanMsg *msg, int32_t count)
{
char *ptr;
unsigned char ch, hex;
int32_t cnt, i;
char *type_str;
char id_str[10];
char dlc_str[3];
char msg_str[30];

for (;count;count--)
  {
  // Message Format
  if ((msg->MsgRTR) && (msg->MsgEFF))
    type_str = "EFF/RTR";
  else if (msg->MsgEFF)
    type_str = "EFF    ";
  else if (msg->MsgRTR)
    type_str = "STD/RTR";
  else
    type_str = "STD    ";
  // ID
  if (msg->MsgEFF)
   safe_sprintf(id_str, 10, "%08X", msg->Id);
  else
   safe_sprintf(id_str, 10, "     %03X", msg->Id);
  // Dlc
  cnt = msg->MsgLen;
  safe_sprintf(dlc_str, 2, "%1u", (uint32_t)cnt);
  // Daten
  if (msg->MsgRTR)
    cnt = 0;
  ptr = msg_str;
  i=0;
  for (; cnt; cnt--)
    {
    ch = msg->MsgData[i++];
    //ptr += sprintf(ptr, "%02x ", ch);
    hex = ch >> 4;
    if (hex > 9)
      *ptr++ = 55 + hex;
    else
      *ptr++ = '0' + hex;
    hex = ch & 0x0F;
    if (hex > 9)
      *ptr++ = 55 + hex;
    else
      *ptr++ = '0' + hex;
    *ptr++ = ' ';
    }
  *ptr = '\0';
  fprintf(LogFile, "%8u.%8u  %s | %s | %s | %s\n", 
          msg->Time.Sec, msg->Time.USec, type_str, id_str, dlc_str, msg_str);
  msg++;
  }
}


/*
******************** LogCanMessages ********************
Funktion  : Eine CAN Message ins Log File schreiben

Eingaben  : msg => CAN Message

Ausgaben  : keine

Call's    : LogWriteMessages, fflush, DRV_LOCK_ENTER, DRV_LOCK_LEAVE
*/
void LogCanMessages(uint32_t log_type, uint32_t log_flags, struct TCanMsg *msg, int32_t count)
{
if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(&LogLock);
  LogWriteMessages(msg, count);
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(&LogLock);
  }
}


/*
******************** LogCanMessage ********************
Funktion  : Eine CAN Message ins Log File schreiben

Eingaben  : msg => CAN Message

Ausgaben  : keine

Call's    : LogWriteMessages, fflush, DRV_LOCK_ENTER, DRV_LOCK_LEAVE
*/
void LogCanMessage(uint32_t log_type, struct TCanMsg *msg, const char *format_str, ...)
{
va_list ap;

if ((LogEnable) && (log_type & LogFlags))
  {
  DRV_LOCK_ENTER(&LogLock);
  va_start(ap, format_str);
  vfprintf(LogFile, format_str, ap);
  va_end(ap);
  LogWriteMessages(msg, 1);
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
  DRV_LOCK_LEAVE(&LogLock);
  }
}


/*
******************** LogStringList ********************
Funktion  :
         [Bezeichner]=[Value];[Bezeichner]=[Value];[..]=[..]
Eingaben  :

Ausgaben  : keine

Call's    : fprintf, fflush, DRV_LOCK_ENTER, DRV_LOCK_LEAVE
*/
void LogStringList(uint32_t log_type, uint32_t log_flags, char *str_list)
{
char *tmp_str, *str, *key, *value;
int32_t res;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(&LogLock);    
  tmp_str = mhs_strdup(str_list);
  str = tmp_str;
  while (*str)
    {
    // **** Key auslesen
    key = get_item_as_string(&str, "=;", &res);
    if ((res < 0) || (!key))
      break;    
    // Fhrende Leerzeichen lschen
    while (*key == ' ')
      key++;
    if (*key == '\0')
      break;    
    // **** Value auslesen
    value = get_item_as_string(&str, "=;", &res);
    if (res == 1) // = gefunden (Key) fehler
      break;
    // Fhrende Leerzeichen lschen
    if (value)
      {
      while (*value == ' ')
        value++;
      if (*value == '\0')
        value = NULL;
      }  
    fprintf(LogFile, "   %s = %s\n", key, value);        
    }
  safe_free(tmp_str);            
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(&LogLock);
  }
}


/*
********************  ********************
Funktion  :

Eingaben  :

Ausgaben  : keine

Call's    : fprintf, fflush, DRV_LOCK_ENTER, DRV_LOCK_LEAVE
*/
void LogSysError(char *file_name, char *func_name, int32_t error_no)
{
if ((!file_name) || (!func_name))
  return;
if (LogEnable)
  {
//  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(&LogLock);
  fprintf(LogFile, "FATAL-ERROR: File:%s, Funktion:%s, Error:%d\n", file_name, func_name, (int32_t)error_no);
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
//  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(&LogLock);
  }
}


/*
******************** LogHexDump ********************
Funktion  :

Eingaben  :

Ausgaben  : keine

Call's    : fprintf, fflush, DRV_LOCK_ENTER, DRV_LOCK_LEAVE
*/
void LogHexDump(uint32_t log_type, uint32_t log_flags, char *data, int32_t count)
{
char str[50];
char *ptr;
int32_t cnt;
unsigned char ch, hex;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(&LogLock);
  while (count)
    {
    if (count > 16)
      cnt = 16;
    else
      cnt = count;
    count -= cnt;
    ptr = str;
    for (;cnt;cnt--)
      {
      ch = (unsigned char)*data++;
      hex = ch >> 4;
      if (hex > 9)
        *ptr++ = 'A' + hex;
      else
        *ptr++ = '0' + hex;
      hex = ch & 0x0F;
      if (hex > 9)
        *ptr++ = 'A' + hex;
      else
        *ptr++ = '0' + hex;
      *ptr++ = ' ';
      }
    *ptr = '\0';
    fprintf(LogFile, "   %s\n", str);
    }
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(&LogLock);
  }
}


/*
******************** LogPrintDrvStatus ********************
Funktion  :

Eingaben  :

Ausgaben  : keine

Call's    : fprintf, fflush, DRV_LOCK_ENTER, DRV_LOCK_LEAVE
*/
void LogPrintDrvStatus(uint32_t log_type, uint32_t log_flags, uint32_t channel, struct TDeviceStatus *dev_status)
{
int32_t drv_status;
unsigned char can_status;
unsigned char fifo_status;

drv_status = dev_status->DrvStatus;
can_status = dev_status->CanStatus;
fifo_status = dev_status->FifoStatus;
if ((LogEnable) && (log_type & LogFlags))
  {
  if (drv_status > DRV_STATUS_MAX)
    drv_status = DRV_STATUS_MAX;
  if (can_status > CAN_STATUS_MAX)
    can_status = CAN_STATUS_MAX;
  if (fifo_status > FIFO_STATUS_MAX)
    fifo_status = FIFO_STATUS_MAX;
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(&LogLock);
  if (can_status & BUS_FAILURE)    
    fprintf(LogFile, "STATUS: Ch %d: Drv=%s, Can=%s [BUS-FAILURE], Fifo=%s\n", channel, DrvStatusStrings[drv_status],
      CanStatusStrings[can_status & 0x0F], FifoStatusStrings[fifo_status]);
  else      
    fprintf(LogFile, "STATUS: Ch %d: Drv=%s, Can=%s, Fifo=%s\n", channel, DrvStatusStrings[drv_status],
      CanStatusStrings[can_status & 0x0F], FifoStatusStrings[fifo_status]);
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(&LogLock);
  }
}


void LogPrintUsbDevices(uint32_t log_type, uint32_t log_flags, struct TCanDevicesList *list, int32_t num_devs)
{
int32_t i;

if (!list)
  num_devs = 0;
if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(&LogLock);
  fprintf(LogFile, "    Open    |Tiny-CAN Id | Snr        | USB-Description \n");
  fprintf(LogFile, "============+============+============+=====================\n");
  if (num_devs <= 0)
    fprintf(LogFile, "keine Devices gefunden.\n");
  else
    {
    for (i = 0; i < num_devs; i++)
      {
      fprintf(LogFile, " 0x%08X | 0x%08X | %10s | %s\n", list[i].TCanIdx, list[i].HwId,
          list[i].SerialNumber, list[i].Description);
      }
    }
  fprintf(LogFile, "============+============+============+=====================\n");
#ifndef LOG_FILE_MODE_BUFFERED
  fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(&LogLock);
  }
}
