/***************************************************************************
                          sm_drv.c  -  description
                             -------------------
    begin             : 23.03.2008
    copyright         : (C) 2008 by MHS-Elektronik GmbH & Co. KG, Germany
    author            : Klaus Demlehner, klaus@mhs-elektronik.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software, you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation.             *
 *                                                                         *
 ***************************************************************************/

/**
    Library to talk to Tiny-CAN devices. You find the latest versions at
       http://www.tiny-can.com/
**/
#include "global.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "com_io.h"
#include "util.h"
#include "sm_drv.h"


/*

Protokoll Aufbau
================

             +---------------------------------------------------+
             |          Kommando an den Mikrocontroller          |
 PC -> uC    |          ===============================          |
             |  - Kommando ohne Parameter, Single Kommando       |
             |  - Kommando mit Parametern, Ext. I / II Kommando  |
             +---------------------------------------------------+
             |           Antworten vom Mikrocontroller           |
             |           =============================           |
 PC <- uC    | - Positive Antwort ohne Parameter (ACK)           |
             | - Negative Antwort mit Fehlercode (NACK)          |
             | - Positive Antwort mit Parameter (ACK)            |
             +---------------------------------------------------+


Kommando ohne Parameter - Single Kommando
=========================================

 Byte |   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   1  |     0     |                                      Kommando                                     |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   2  |                                      8 Bit CRC Prfsumme                                      |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+


Kommando mit max. 255 Zeichen Parameter - Ext. I Kommando
=========================================================

 Byte |   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   1  |     1     |     0     |                                Kommando                               |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   2  |                                        Anzahl Parameter                                       |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   3  |                                             Daten                                             |
      +-----------                                                                         -----------+
   .. |                                            . . . .                                            |
      +-----------                                                                         -----------+
   n  |                                            . . . .                                            |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
  n+1 |                                      8 Bit CRC Prfsumme                                      |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+


Kommando mit max. 16383 Zeichen Parameter - Ext. II Kommando
============================================================

 Byte |   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   1  |     1     |     1     |                                Kommando                               |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   2  |     0     |     0     |                Anzahl Parameter High Byte (Bit 13 - 8)                |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   3  |                             Anzahl Parameter Low Byte (Bit 7 - 0)                             |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   4  |                                             Daten                                             |
      +-----------                                                                         -----------+
   .. |                                            . . . .                                            |
      +-----------                                                                         -----------+
   n  |                                            . . . .                                            |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
  n+1 |                                      8 Bit CRC Prfsumme                                      |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+



Antworten - ACKs
================

Positiv acknowledge - ACK
=========================

 Byte |   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   1  |     1     |     0     |     0     |     1     |     0     |     1     |     0     |     1     |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+

Es wird keine Prfsumme gesendet


Negativ acknowledge - NACK
==========================

 Byte |   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   1  |     1     |     0     |     1     |                         Fehlercode                        |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+

Es wird keine Prfsumme gesendet


Positiv acknowledge mit max. 16383 Parametern - ACK
===================================================

 Byte |   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   1  |     0     |     1     |                Anzahl Parameter High Byte (Bit 13 - 8)                |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   2  |                             Anzahl Parameter Low Byte (Bit 7 - 0)                             |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
   3  |                                             Daten                                             |
      +-----------                                                                         -----------+
   .. |                                            . . . .                                            |
      +-----------                                                                         -----------+
   n  |                                            . . . .                                            |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
  n+1 |                                      8 Bit CRC Prfsumme                                      |
      +-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+

*/

static unsigned char CRCTabelle[] = {
    0x00, 0x5e, 0xbc, 0xe2, 0x61, 0x3f, 0xdd, 0x83,
    0xc2, 0x9c, 0x7e, 0x20, 0xa3, 0xfd, 0x1f, 0x41,
    0x9d, 0xc3, 0x21, 0x7f, 0xfc, 0xa2, 0x40, 0x1e,   // 10
    0x5f, 0x01, 0xe3, 0xbd, 0x3e, 0x60, 0x82, 0xdc,
    0x23, 0x7d, 0x9f, 0xc1, 0x42, 0x1c, 0xfe, 0xa0,   // 20
    0xe1, 0xbf, 0x5d, 0x03, 0x80, 0xde, 0x3c, 0x62,
    0xbe, 0xe0, 0x02, 0x5c, 0xdf, 0x81, 0x63, 0x3d,   // 30
    0x7c, 0x22, 0xc0, 0x9e, 0x1d, 0x43, 0xa1, 0xff,
    0x46, 0x18, 0xfa, 0xa4, 0x27, 0x79, 0x9b, 0xc5,   // 40
    0x84, 0xda, 0x38, 0x66, 0xe5, 0xbb, 0x59, 0x07,
    0xdb, 0x85, 0x67, 0x39, 0xba, 0xe4, 0x06, 0x58,   // 50
    0x19, 0x47, 0xa5, 0xfb, 0x78, 0x26, 0xc4, 0x9a,
    0x65, 0x3b, 0xd9, 0x87, 0x04, 0x5a, 0xb8, 0xe6,   // 60
    0xa7, 0xf9, 0x1b, 0x45, 0xc6, 0x98, 0x7a, 0x24,
    0xf8, 0xa6, 0x44, 0x1a, 0x99, 0xc7, 0x25, 0x7b,   // 70
    0x3a, 0x64, 0x86, 0xd8, 0x5b, 0x05, 0xe7, 0xb9,
    0x8c, 0xd2, 0x30, 0x6e, 0xed, 0xb3, 0x51, 0x0f,   // 80
    0x4e, 0x10, 0xf2, 0xac, 0x2f, 0x71, 0x93, 0xcd,
    0x11, 0x4f, 0xad, 0xf3, 0x70, 0x2e, 0xcc, 0x92,   // 90
    0xd3, 0x8d, 0x6f, 0x31, 0xb2, 0xec, 0x0e, 0x50,
    0xaf, 0xf1, 0x13, 0x4d, 0xce, 0x90, 0x72, 0x2c,   // A0
    0x6d, 0x33, 0xd1, 0x8f, 0x0c, 0x52, 0xb0, 0xee,
    0x32, 0x6c, 0x8e, 0xd0, 0x53, 0x0d, 0xef, 0xb1,   // B0
    0xf0, 0xae, 0x4c, 0x12, 0x91, 0xcf, 0x2d, 0x73,
    0xca, 0x94, 0x76, 0x28, 0xab, 0xf5, 0x17, 0x49,   // C0
    0x08, 0x56, 0xb4, 0xea, 0x69, 0x37, 0xd5, 0x8b,
    0x57, 0x09, 0xeb, 0xb5, 0x36, 0x68, 0x8a, 0xd4,   // D0
    0x95, 0xcb, 0x29, 0x77, 0xf4, 0xaa, 0x48, 0x16,
    0xe9, 0xb7, 0x55, 0x0b, 0x88, 0xd6, 0x34, 0x6a,   // E0
    0x2b, 0x75, 0x97, 0xc9, 0x4a, 0x14, 0xf6, 0xa8,
    0x74, 0x2a, 0xc8, 0x96, 0x15, 0x4b, 0xa9, 0xf7,   // F0
    0xb6, 0xe8, 0x0a, 0x54, 0xd7, 0x89, 0x6b, 0x35};


#define updcrc8(cp, crc) (CRCTabelle[crc ^ cp])


int32_t SMDrvClear(TCanDevice *dev)
{
int i;
unsigned char c;

if (ComIsOpen(dev->Io) < 1)          // Com Port geffnet ?
  return(ERR_SM_PORT_NOT_OPEN);
ComFlushBuffer(dev->Io);             // Sende und Empfangspuffer lschen
for (i = 10; i; i--)
  {
  mhs_sleep(10);
  if (ComReadByte(dev->Io, &c) < 1)
    return(0);
  ComFlushBuffer(dev->Io);           // Sende und Empfangspuffer lschen
  }
return(-1);
}


int32_t SMDrvCommando(TCanDevice *dev, uint32_t sleep_time)
{
unsigned char c, crc, read_crc;
uint16_t i, param_cnt;
uint32_t tx_cnt;
unsigned char *tx_puf;
int32_t res;

tx_puf = &dev->TxDDataPuffer[0];
if (ComIsOpen(dev->Io) < 1)          // Com Port geffnet ?
  return(ERR_SM_PORT_NOT_OPEN);

c = dev->Kommando;
// **** Comando senden
*tx_puf++ = c;
crc = CRCTabelle[c];
c &= 0xC0;                       // Commando Type -> Single, Ext I oder II
// **** Parameter lnge senden
if (c == 0x80)                   // Ext. I
  {
  param_cnt = (uint16_t)dev->TxDParameter.Count;
  tx_cnt = param_cnt + 3;
  c = (unsigned char)param_cnt;
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  }
else if (c == 0xC0)              // Ext. II
  {
  param_cnt = (uint16_t)dev->TxDParameter.Count;
  tx_cnt = param_cnt + 4;
  c = (unsigned char)((param_cnt >> 8) & 0x3F);
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  c = (unsigned char)param_cnt;
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  }
else
  {
  param_cnt = 0;
  tx_cnt = 2;
  }
// **** Parameter senden
for (i = 0; i < param_cnt; i++)
  {
  c = dev->TxDParameter.Data[i];
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  }
// **** CRC senden
crc = CRCTabelle[CRCTabelle[crc]];
*tx_puf = crc;
if (ComWriteData(dev->Io ,(char *)dev->TxDDataPuffer, tx_cnt) < 0)
  {
  ComClose(dev->Io);
  return(ERR_SM_TX_CMD);
  }
if (sleep_time)  
  mhs_sleep(sleep_time);  
// **** ACK Empfangen
dev->RxDParameter.Count = 0;
// **** ACK empfangen
if ((res = ComReadByte(dev->Io, &c)) < 0)  // ACK Byte Empfangen
  {
  ComClose(dev->Io);
  return(ERR_SM_RX_ACK);
  }
if (res == 0)
  return(ERR_SM_RX_ACK_TIMEOUT);
dev->ACKByte = c;
crc = CRCTabelle[c];
if ((c & 0xC0) == 0x40)          // noch Parameter Empf. ? (ACK Bit 7 = 0 Bit 6 = 1)
  {                              // Ja =>
  param_cnt = ((uint16_t)c << 8) & 0x3F00;   // Anzahl Parameter (High Byte)
  if ((res = ComReadByte(dev->Io, &c)) < 0)  // Anzahl Parameter (Low Byte) lesen
    {
    ComClose(dev->Io);
    return(ERR_SM_RX_PARAM_CNT);
    }
  if (res == 0)
    return(ERR_SM_RX_PARAM_CNT_TIMEOUT);
  crc = updcrc8(c, crc);
  param_cnt |= c;
  dev->RxDParameter.Count = param_cnt;
  // Parameter lesen
  if ((res = ComReadData(dev->Io, (char *)dev->RxDParameter.Data, param_cnt)) < 0)
    {
    ComClose(dev->Io);
    return(ERR_SM_RX_PARAM);
    }
  if (res == 0)
    return(ERR_SM_RX_PARAM_TIMEOUT);
  if ((uint32_t)res != param_cnt)   // Anzahl Empfangener Parameter stimmt nicht
    return(ERR_SM_RX_PARAM_COUNT);
  for (i = 0; i < param_cnt; i++)
    {
    c = dev->RxDParameter.Data[i];
    crc = updcrc8(c, crc);
    }
  // CRC lesen
  if ((res = ComReadByte(dev->Io, &read_crc)) < 0)
    {
    ComClose(dev->Io);
    return(ERR_SM_RX_CRC);
    }
  if (res == 0)
    return(ERR_SM_RX_CRC_TIMEOUT);
  crc = CRCTabelle[CRCTabelle[crc]];
  if (crc != read_crc)           // CRC prfen
    return(ERR_SM_WRONG_CRC);
  }
else
  {
  if ((c & 0xE0) == 0xA0)
    return(ERR_SM_NACK);                  // NACK
  else if (c != 0x95)
    return(ERR_SM_WRONG_ACK);             // Fehler beim ACK Empfang
  }
return(0);
}
