/***************************************************************************
                          tar_drv.c  -  description
                             -------------------
    begin             : 01.11.2010
    copyright         : (C) 2010 by MHS-Elektronik GmbH & Co. KG, Germany
    author            : Klaus Demlehner, klaus@mhs-elektronik.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software, you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation.             *
 *                                                                         *
 ***************************************************************************/

/**
    Library to talk to Tiny-CAN devices. You find the latest versions at
       http://www.tiny-can.com/
**/
#include "config.h"
#include "global.h"
#include <stdio.h>
#include <string.h>
#include "util.h"
#include "errors.h"
#include "com_io.h"
#include "sm_drv.h"
#include "log.h"
#include "tcan_com.h"
#include "tar_drv.h"



/*

Status:
=======

|   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
+-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
| CAN 0 TxD |   CAN 0 TxD Status    | CAN 0 RxD | CAN 0 RxD |           CAN 0 Status            |
|   Error   |                       |     OV    |   Pend    |                                   |



CAN x Status:

  0 = Controller Ok
  1 = Error warning
  2 = Error passiv
  3 = Bus Off
  4 = Empfangs Puffer überlauf
  5 = CAN Bus Stop
  7 = Power Up

CAN x TxD Status:
  00 = Empty
  01 = Pend
  10 = Limit ( 1 / 2 Puffergröße überschritten)
  11 = OV

CAN0TxStatus:
=============

|   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
+-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
|   Box 8   |   Box 7   |   Box 6   |   Box 5   |   Box 4   |   Box 3   |   Box 2   |   Box 1   |
|   Error   |   Error   |   Error   |   Error   |   Error   |   Error   |   Error   |   Error   |


CAN1TxStatus:
=============

|   Bit 7   |   Bit 6   |   Bit 5   |   Bit 4   |   Bit 3   |   Bit 2   |   Bit 1   |   Bit 0   |
+-----------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+
|   Box 8   |   Box 7   |   Box 6   |   Box 5   |   Box 4   |   Box 3   |   Box 2   |   Box 1   |
|   Error   |   Error   |   Error   |   Error   |   Error   |   Error   |   Error   |   Error   |

*/

static const uint32_t CAN_SPEED_TAB[] = {10,     // 0 = 10 kBit/s
                                         20,     // 1 = 20 kBit/s
                                         50,     // 2 = 50 kBit/s
                                         100,    // 3 = 100 kBit/s
                                         125,    // 4 = 125 kBit/s
                                         250,    // 5 = 250 kBit/s
                                         500,    // 6 = 500 kBit/s
                                         800,    // 7 = 800 kBit/s
                                         1000,   // 8 = 1 MBit/s
                                         0};     // Ende der Liste


#define SetUInt32ToParam(p, i, d) do { \
  (p)[(i)] = (unsigned char)(d); \
  (p)[(i+1)] = (unsigned char)((d) >> 8); \
  (p)[(i+2)] = (unsigned char)((d) >> 16); \
  (p)[(i+3)] = (unsigned char)((d) >> 24); \
  } while(0)

/**************************************************************/
/* Tiny-CAN Status abfragen                                   */
/**************************************************************/
int32_t TARGetStatus(TCanDevice *dev, struct TTarStatus *tar_status)
{
int32_t err;

tar_status->Status.Byte = 0;
tar_status->ExStatus.Byte = 0;
dev->Kommando = CMD_GET_STATUS_ALL;
err = SMDrvCommando(dev, 0);
if (dev->RxDParameter.Count < dev->ModulDesc->StatusSize)
  err = ERR_SM_DRV_ACK_COUNT;
else
  {
  if (dev->ModulDesc->StatusSize > 1)
    tar_status->ExStatus.Byte = dev->RxDParameter.Data[1];
  tar_status->Status.Byte = dev->RxDParameter.Data[0];
  }
LOG_FUNCTION_ERROR("TARGetStatus", err);
return(err);
} 


/**************************************************************/
/* Empfangs-Filter setzten                                    */
/**************************************************************/
int32_t TARSetFilter(TCanDevice *dev, uint32_t index, struct TMsgFilter *msg_filter)
{
int32_t err;
unsigned char filter_index;

err = 0;
filter_index = (unsigned char)(index & 0xFF) - 1;
if (filter_index > 7)
  return(ERR_INDEX_RANGE);
dev->TxDParameter.Data[0] = filter_index;
if (msg_filter->FilEnable)
  {
  if (msg_filter->FilEFF)
    dev->TxDParameter.Data[1] = 0xC0;
  else
    dev->TxDParameter.Data[1] = 0x80;
  SetUInt32ToParam(dev->TxDParameter.Data, 2, msg_filter->Maske);
  SetUInt32ToParam(dev->TxDParameter.Data, 6, msg_filter->Code);
  dev->TxDParameter.Count = 10;
  }
else
  {
  dev->TxDParameter.Data[1] = 0x00;
  dev->TxDParameter.Count = 2;
  }
dev->Kommando = CMD_CAN0_RXD_SET_FILTER;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARSetFilter", err);
return(err);
}


/**************************************************************/
/* Ping / Pong                                                */
/**************************************************************/
int32_t TARCommSync(TCanDevice *dev)   // <*> neu
{
int32_t err, res, i;
unsigned char c;

if (ComIsOpen(dev->Io) < 1)
  return(ERR_SM_PORT_NOT_OPEN);
err = ERR_SM_RX_ACK_TIMEOUT;  
for (i = 3; i; i--)
  {
  if (ComWriteByte(dev->Io, 0x00) < 0)
    {
    err = ERR_SM_TX_DATA;
    break;
    }
  if ((res = ComReadByte(dev->Io, &c)) < 0)
    {
    err = ERR_SM_RX_DATA;
    break;
    }
  if ((res == 1) && (c == 0x00))      
    return(0);    
  mhs_sleep(150);
  ComFlushBuffer(dev->Io);
  }
LOG_FUNCTION_ERROR("TARCommSync", err);  
return(err);
}


int32_t TARPing(TCanDevice *dev)   // <*> neu
{
int32_t err;

dev->Kommando = CMD_PING;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARPing", err);
return(err);
}


/*int32_t TARPing(TCanDevice *dev)   // <*> alt
{
int32_t err, res, i;
unsigned char c;

if (ComIsOpen(dev->Io) < 1)
  return(ERR_SM_PORT_NOT_OPEN);
err = ERR_SM_RX_ACK_TIMEOUT;  
for (i = 3; i; i--)
  {
  if (ComWriteByte(dev->Io, 0x00) < 0)
    {
    err = ERR_SM_TX_DATA;
    break;
    }
  if ((res = ComReadByte(dev->Io, &c)) < 0)
    {
    err = ERR_SM_RX_DATA;
    break;
    }
  if ((res == 1) && (c == 0x00))
    {
    dev->Kommando = CMD_PING;
    if (!SMDrvCommando(dev, 0))
      return(0);
    }
  mhs_sleep(150);
  ComFlushBuffer(dev->Io);
  }
return(err);
} */


int32_t TARInfoStartRead(TCanDevice *dev)
{
int32_t err;

dev->Kommando = CMD_START_READ_INFO;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARInfoStartRead", err);
return(err);
}


int32_t TARInfoGetNext(TCanDevice *dev, struct TTarInfoVar *info)
{
int32_t err;

dev->Kommando = CMD_READ_INFO;
if (!(err = SMDrvCommando(dev, 0)))
  {
  if (dev->RxDParameter.Count == 0)
    return(0);
  else if (dev->RxDParameter.Count > sizeof(struct TInfoVar))
    err = ERR_SM_DRV_ACK_COUNT;
  else
    {
    memcpy(info, &dev->RxDParameter.Data[0], dev->RxDParameter.Count);
    return(1);
    }
  }
LOG_FUNCTION_ERROR("TARInfoGetNext", err);
return(err);
}


/**************************************************************/
/* Hardware TimeStamps aktivieren                             */
/**************************************************************/
int32_t TARSetTimeStamp(TCanDevice *dev, unsigned char enabled)
{
int32_t err;

dev->TxDParameter.Count = 1;
dev->TxDParameter.Data[0] = enabled;
dev->Kommando = CMD_SET_TIMESTAMP;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARSetTimeStamp", err);
return(err);
}


/**************************************************************/
/* CAN Arbeitsmodus einstellen                                */
/**************************************************************/
int32_t TARSetCANMode(TCanDevice *dev, unsigned char can_mode)
{
int32_t err;

dev->TxDParameter.Count = 1;
dev->TxDParameter.Data[0] = can_mode;
dev->Kommando = CMD_CAN0_SET_MODE;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARSetCANMode", err);
return(err);
}


/**************************************************************/
/* CAN Speed einstellen                                       */
/**************************************************************/
int32_t TARCanSetSpeed(TCanDevice *dev, uint32_t speed)
{
uint32_t s;
int32_t err;
unsigned char i;

for (i = 0; (s = CAN_SPEED_TAB[i]) != speed; i++)
  {
  if (!s)
    break;
  }
if (!s)
  err = -1;
else
  {
  i++;
  // **** Speed Einstellen
  dev->TxDParameter.Count = 2;
  dev->TxDParameter.Data[0] = CAN_SPEED_INDEX;  // Index
  dev->TxDParameter.Data[1] = i;
  dev->Kommando = CMD_CAN0_SET;
  err = SMDrvCommando(dev, 0);
  }
LOG_FUNCTION_ERROR("TARCanSetSpeed", err);
return(err);
}


/**************************************************************/
/* CAN TX ACK einstellen                                      */
/**************************************************************/
int32_t TARCanSetTxAckEnable(TCanDevice *dev, unsigned char enable)
{
int32_t err;

// **** Speed Einstellen
dev->TxDParameter.Count = 2;
dev->TxDParameter.Data[0] = CAN_TX_ACK_ENABLE;  // Index
dev->TxDParameter.Data[1] = enable;
dev->Kommando = CMD_CAN0_SET;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARCanSetTxAckEnable", err);
return(err);
}


/**************************************************************/
/* CAN LOG ERRORS einstellen                                  */
/**************************************************************/
int32_t TARCanSetErrorLogEnable(TCanDevice *dev, unsigned char enable)
{
int32_t err;

// **** Speed Einstellen
dev->TxDParameter.Count = 2;
dev->TxDParameter.Data[0] = CAN_ERROR_LOG_ENABLE;  // Index
dev->TxDParameter.Data[1] = enable;
dev->Kommando = CMD_CAN0_SET;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARCanSetErrorLogEnable", err);
return(err);
}


/**************************************************************/
/* CAN Speed über BTR Register einstellen                     */
/**************************************************************/
int32_t TARCanSetBtr(TCanDevice *dev, uint32_t btr)
{
int32_t err;

// **** Speed Einstellen
dev->TxDParameter.Count = 5;
dev->TxDParameter.Data[0] = CAN_SPEED_BTR;  // Index
SetUInt32ToParam(dev->TxDParameter.Data, 1, btr);
dev->Kommando = CMD_CAN0_SET;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARCanSetBtr", err);
return(err);
}


/**************************************************************/
/* CAN TxD Boxes sende Intervall einstellen                   */
/**************************************************************/
int32_t TARCanSetIntervall(TCanDevice *dev, uint32_t index, uint32_t intervall)
{
int32_t err;
unsigned char box_index;

if (intervall)
  {
  if (intervall < 1000)
    intervall = 1;
  else
    intervall = intervall / 1000;
  }
box_index = (unsigned char)(index & 0xFF) - 1;
if (box_index > 15)
  return(ERR_INDEX_RANGE);
dev->TxDParameter.Count = 3;
dev->TxDParameter.Data[0] = box_index;
dev->TxDParameter.Data[1] = lo(intervall);
dev->TxDParameter.Data[2] = hi(intervall);
dev->Kommando = CMD_CAN0_TXD_IDX_SET_TIMER;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARCanSetIntervall", err);
return(err);
}


/**************************************************************/
/* CAN Controller Reset                                       */
/**************************************************************/
int32_t TARCANReset(TCanDevice *dev)
{
int32_t err;

dev->TxDParameter.Count = 1;
dev->TxDParameter.Data[0] = 0x30;
dev->Kommando = CMD_CAN0_SET_MODE;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARCANReset", err);
return(err);
}


/**************************************************************/
/* Empfangs - Fifo löschen                                    */
/**************************************************************/
int32_t TARRxFifoClear(TCanDevice *dev, uint32_t index)
{
int32_t err;
unsigned char fifo_idx;

dev->TxDParameter.Count = 0;
fifo_idx = (unsigned char)(index & 0x000000FFL);
if (!fifo_idx)
  dev->Kommando = CMD_CAN0_RXD_CLEAR;
else
  {
  dev->TxDParameter.Count = 1;
  dev->TxDParameter.Data[0] = fifo_idx;
  dev->Kommando = CMD_CAN0_RXD_IDX_CLEAR;
  }
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARRxFifoClear", err);
return(err);
}


/**************************************************************/
/* Sende - Fifo Daten lesen                                   */
/**************************************************************/
int32_t TARTxFifoPut(TCanDevice *dev, uint32_t index, struct TCanMsg *msgs, uint16_t count)
{
int32_t err;
uint16_t cnt;
unsigned char fifo_idx, len;
unsigned char *dest;

err = 0;
fifo_idx = (unsigned char)(index & 0x000000FFL);
if (!fifo_idx)
  {
  dest = &dev->TxDParameter.Data[0];
  cnt = 0;
  }
else
  {
  dev->TxDParameter.Data[0] = fifo_idx;
  dest = &dev->TxDParameter.Data[1];
  cnt = 1;
  }

for (; count; count--)
  {
  len = msgs->MsgLen;
  if (msgs->MsgEFF)
    {
    if (!msgs->MsgRTR)
      *dest++ = len | 0xC0;  // EXT
    else
      {
      *dest++ = len | 0xE0;  // EXT / RTR
      len = 0;
      }
    *(uint32_t *)dest = (uint32_t)msgs->Id;
    dest += 4;
    if (len)
      {
      memcpy(dest, &msgs->MsgData, len);
      dest += len;
      }
    len += 5;
    }
  else
    {
    if (!msgs->MsgRTR)
      {  // STD
      *dest++ = (unsigned char)(((msgs->Id >> 4) & 0x70) | len);
      *dest++ = (unsigned char)msgs->Id;
      memcpy(dest, &msgs->MsgData, len);
      dest += len;
      len += 2;
      }
    else
      {  // STD / RTR
      *dest++ = len | 0xA0;
      *(uint16_t *)dest = (uint16_t)msgs->Id;
      dest += 2;
      len = 3;
      }
    }
  cnt += len;
  msgs++;
  }

dev->TxDParameter.Count = cnt;
if (!fifo_idx)
  dev->Kommando = CMD_CAN0_TXD;
else
  dev->Kommando = CMD_CAN0_TXD_IDX;
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARTxFifoPut", err);
return(err);
}


/**************************************************************/
/* Sende - Fifo löschen                                       */
/**************************************************************/
int32_t TARTxFifoClear(TCanDevice *dev, uint32_t index)
{
int32_t err;
unsigned char fifo_idx;

dev->TxDParameter.Count = 0;
fifo_idx = (unsigned char)(index & 0x000000FFL);
if (!fifo_idx)
  dev->Kommando = CMD_CAN0_TXD_CLEAR;
else
  {
  dev->TxDParameter.Count = 1;
  dev->TxDParameter.Data[0] = fifo_idx;
  dev->Kommando = CMD_CAN0_TXD_IDX_CLEAR;
  }
err = SMDrvCommando(dev, 0);
LOG_FUNCTION_ERROR("TARTxFifoClear", err);
return(err);
}

